	.area code (rel)
;       include buffers.mac             ;all buffer-related stuff in here
;					;plus all (eventually) variables
;
; the buffer list is kept from "bottom" to the end of ram.  the format of the
; buffers is:
;+------+--------+-------+---------------------------------------------------+
;| next | nbytes | nread | data						     |
;+------+--------+-------+---------------------------------------------------+
;
; 2 bytes 1 byte   1 byte  124 bytes  (total 128 bytes)
; next	 pointer to next buffer on this buffer chain (or 0 if no more)
; nbytes number of bytes in this buffer that are valid
; nread  number of bytes read from this buffer (used by getchar)
; data   124 bytes of data (not all is necessarily valid, see nbytes field)
;
; the buffer pool is all here, and as processes need buffer space, it is all
; allocated out of this pool.  see allo_buf and free_buffer code.


;---------------------------------------------------------------------------
; return in hl a pointer to a free buffer.  if there are not more buffers,
; return with z flag set.
; destroys no registers except return value hl.
; is called from an interrupt routine, so this operation is atomic.

	.globl allo_buf
allo_buf:

	push	bc
	push	af

	ld	hl,(free)		; get pointer to head of free list
	ld	a,h
	or	l
        jr      nz,ok_allo_buf          ; assure we're not off the end

;get here if no more buffers.  return z set - do not disturb a.
	pop	af
	ld	b,a			; tuck a away for a moment...
	xor	a			; turn on z bit
	ld	a,b			; retreive original a
	pop	bc
	ret

ok_allo_buf:

	xor	a
        ld      c,(hl)                  ; grab lo byte of next free buffer
	ld	(hl),a			; clear it out
	inc	hl
	ld	b,(hl)			; "ld bc,(hl)" now hi byte
	ld	(hl),a			; clear it out, too
	ld	(free),bc		; update with new free list pointer

	dec	hl			; now hl is at head of new buffer

	pop	af
	ld	b,a			; tuck a away for a moment...
	ld	a,#1
	or	a			; turn z bit off (i.e., all ok)
	ld	a,b			; retreive original a

	pop	bc
	ret

;---------------------------------------------------------------------------
; free_buffer gets passed a pointer (in hl) to a buffer to be freed.  the
; buffer is placed on the head of the free list.  the nbytes & nread fields
; are made 0 before placing on free list.
; this routine is called at interrupt level, so results are atomic.
; no registers are disturbed at all.  the free pointer is updated, however.
; 159 t states [ 63.6 usec @ 2.5 mhz ]

	.globl free_buffer
free_buffer:
	push	af
	push	bc		;we'll use these
	push	hl		;this will be new head of free list

	ld	bc,(free)	;get old free head
	ld	(hl),c		;put on free chain, first low byte...
	inc	hl
	ld	(hl),b		; ...now hi byte
	xor	a
	inc	hl
	ld	(hl),a		; zero out nbytes field
	inc	hl
	ld	(hl),a		; and the nread field of new head of free

	pop	hl		;get new head of free list back
	ld	(free),hl	;and save it in memory where it belongs

	pop	bc
	pop	af
	ret
; --------------------------------------------------------------------------
; putchar - hl contains pointer to buffer, a contains the character to put
; into the buffer.  upon return, char is put into this buffer if there is
; room, else another buffer is allocated and hl is updated to point to this
; new buffer.  the new buffer is chained onto the old buffer in this case.
; the calling routine is responsible for maintaing both the head of a
; particular buffer chain (if it needs it), and the current buffer being
; manipulated. this routine is called at interrupt level, so is atomic.  no
; registers disturbed, except that hl may have a new value.
; 211 t states [  84.4 usec @ 2.5 mhz ]	no new buffer required
; 338 t states [ 135.2 usec @ 2.5 mhz ]	new buffer needed

; [jdz] if carry flag is on, then memory is not enough to putchar.

	.globl putchar
putchar:
	push	bc
	push	ix
	push	af
	push	hl		;do it this way for a reason...

	pop	ix		;get buffer pointer into ix
	ld	a,2(ix)		;grab nbytes field
	cp	#124		;max of 124 chars in a buffer
        jr      nz,putc1
        call    put_ned_buf
; if it takes this call, it returns with a new buffer, with hl pointing to
; it (as well as ix), and with a reg set to 0.
; else just plunk into buffer
        jr      nc,putc1
	pop	af
	scf
        jr      putc2
putc1:
	inc	2(ix)		;one more char will go into this buffer
	ld	c,a		;get previous nbytes
	xor	a
	ld	b,a		; bc <- nbytes, filled out to 16 bits
	add	ix,bc		; update ix to point to where char goes
	pop	af		; retreive the char we want to save
	ld	4(ix),a		; save it in this buffer
	or	a		; reset cy flag.
putc2:
	pop	ix
	pop	bc
	ret			;done for the moment

; 127 t states [ 50.8 usec @ 2.5 mhz ] (really part of prev routine)
put_ned_buf:			; prev buffer filled, get a new one
	push	de		; working registers
	push	hl		; save current buffer pointer
        call    allo_buf	; grab a new buffer, addr is in hl
        jr      nz,putcnb1
	pop	hl
	scf
        jr      putcnb2
putcnb1:
	ex	de,hl		; "ld de,hl" - get new addr into de for now
	pop	hl
	ld	(hl),e		; link new buffer onto chain, lo byte first
	inc	hl
	ld	(hl),d		; now hi byte, chaining done

	ex	de,hl		; update hl for orig. calling routine's use
	push	hl
	pop	ix		; upper routine needs ix pointing to new buf
	xor	a		; and a is nbytes in calling routine, make..
				; zero for a new buffer
putcnb2:
	pop	de		; done with this working register
	ret			; all done here, let calling routine finish

; --------------------------------------------------------------------------
; getchar - grab a character from the buffer pointed at by hl, return in a.
; if the "nread" field of this buf = "nbytes" then this buffer is exhausted,
; so follow the chain on to the next buffer & release old buffer.  if the
; next chain is 0, or if the nbytes field is >= nread field, then there are
; no more bytes.  in this case, return with z bit set; normally return with
; z bit reset (that is, non-zero) indicating a valid char is in a.  note
; that if we need to follow the chain to a new buffer, hl will be updated,
; too, so that the calling routine needs to deal with this.
; 		no registers changed except af and possibly hl.
; called at interrupt level, so operation is atomic.
; 212 t states [  84.8 usec @ 2.5 mhz ]	no new buffer needed
; 493 t states [ 197.2 usec @ 2.5 mhz ]	if following chain

	.globl getchar
getchar:
	push	ix		; save because is working reg
	push	bc		; working regs here

	push	hl
	pop	ix		; ix points to this buffer

	ld	a,3(ix)		; grab nread
	cp	2(ix)		; compare with nbytes
	call	z,getc_new_buf	; if they are same, this buffer is spent

	inc	3(ix)		; we are reading one more char, update nread
	inc	a
	cp	2(ix)
        jr      nz,get_pl_char  ; if not looking at last character

; else, is the "next" pointer 0?
	push	hl
	ld	b,a		; !!!!! save  a   reg  !!!!!!! 4 jan 87
	ld	a,(hl)
	inc	hl
	or	(hl)
	ld	a,b		; !!!! restore a reg  (gasp!)
	pop	hl
        jr      nz,get_pl_char

; else next is 0 and we are on last char - flush it & quit
	call	free_buffer
	pop	bc
	pop	ix
	ret			; note that z bit is set (from above)

; else we can just pluck a character out of this buffer
get_pl_char:
	dec	a		; fix a from above mucking around...

	ld	c,a		; get old nread into bc
	ld	b,#0		; ditto
	add	ix,bc		; fix buffer pointer
	ld	a,#1
	or	a		; make z bit reset
	ld	a,4(ix)		; get the desired byte

	pop	bc
	pop	ix
	ret			; all for this simple case

; old buffer is spent, get new one (if any)

getc_new_buf:
	push	de		; need this reg here
	ld	e,(hl)		; get lo byte of next pointer
	inc	hl
	ld	d,(hl)		; hi byte of next pointer (now all in de)
	dec	hl		; hl now back to point at spent buffer
	call	free_buffer	; give the buffer back

	ex	de,hl		; "ld hl,de" - follow chain
	push	hl
	pop	ix		; init new ix (same as hl in this routine)
	xor	a		; a holds nread (needed above)
	pop	de		; release de from use by this excursion
	ret

; --------------------------------------------------------------------------
; free_chain - must be called from interrupt routine to guarantee
; atomicity.  takes buffer chain pointed at by hl and returns them to free
; buffer list
; 303 t states + (n_on_chain-1)*238 t states
; [ 121.2 usec + (n_on_chain-1)*95.2 usec ]

	.globl free_chain
free_chain:
	push	af
	push	de
	push	hl		; we will muck with these

fc_0:
	ld	e,(hl)		; get lo part of next buffer pointer
	inc	hl
	ld	d,(hl)		; now hi part of next buffer pointer
	dec	hl
	call	free_buffer	; release this buffer
	ld	a,d
	or	e
        jr      z,fc_9          ; if "next" address is 0, we are at end
; else we've got more on this chain - deal with them.
	ex	de,hl		; "ld hl,de" - hl points to "next"
        jr      fc_0

fc_9:
	pop	hl
	pop	de
	pop	af
	ret

; --------------------------------------------------------------------------
; out_queue_ins - places the just-received buffer on the output queue.
; the address of the rx buffer just received is in hl.
; the output queue is a circular buffer.  the output routine keeps sending
; out buffers until its out_head_cbuf pointer equals its out_tail_cbuf
; pointer. the output routine never mucks with the out_tail_cbuf pointer;
; similarly, this routine never changes the out_head_cbuf pointer.  so it
; is possible to
; insert new entries into the output circular buffer queue without
; disturbing the entry which is being sent to the output port.

	.globl out_queue_ins
out_queue_ins:
	push	af
	push	de
	push	hl		; use these

	ex	de,hl		; "ld de,hl" - move buffer to link addr
	ld	hl,(out_tail_cbuf) ; grab next free location 
	ld	(hl),e		; set lo addr 1st
	inc	hl
	ld	(hl),d		; now hi addr
	inc	hl		; now hl points to next free entry in...
	ld	de,#out_bottom	; ...circ buf, unless we're at end
	or	a		; clear carry
        push    hl              ; (may be needed address)
	sbc	hl,de
	pop	hl		; get back what we think is good
        jr      nz,oqi_0
	
	ld	hl,#out_top	; get here if we're at end of circ buffer.
oqi_0:
	ld	(out_tail_cbuf),hl
	pop	hl
	pop	de
	pop	af		; keep clean
	ret


;---------------------------------------------------------------------------
; tx_queue_ins - similar to out_queue_ins, but with different queue.
; also, increments the byte tx_outstng (which counts the number of
; frames ready to be dumped to the modem port).  this routine, like
; out_queue_ins, does not need to worry about queue wrap-around because
; there are more entries in each of these queues than there are buffers
; available.  yes, i know this is a hack, and wastes some ram space, but it
; means i don't have to check for overflows here.
; the queue is circular, and sometimes i call it a "cbuf" - circular buffer

	.globl tx_queue_ins
tx_queue_ins:
	push	af
	push	de
	push	hl
	ex	de,hl			; "ld de,hl" - save chain head in de
	ld	hl,(tx_tail_cbuf)	; next free location in tx cbuf
	ld	(hl),e
	inc	hl
	ld	(hl),d			; put this chain into tx queue
	inc	hl			; hl is next availble tx queue ...
	ld	de,#tx_bottom		; ... unless we are at bottom of ...
	or	a			; ... the tx queue
	push	hl
	sbc	hl,de
	pop	hl
        jr      nz,tqi_0                ; go there if not at buffer bottom

	ld	hl,#tx_top		; else reload with top of queue val
tqi_0:
	ld	(tx_tail_cbuf),hl	; save next free queue slot
        ld      hl,#tx_outstng
	inc	(hl)			; +1 more frame outstanding now
	pop	hl
	pop	de
	pop	af
	ret

;-----------------------------------------------------------------------------
; setup hl & tx_chain_head for transmission of next chain.

	.globl txnext_cbuf
txnext_cbuf:
	push	af
	push	de
	ld	hl,(tx_head_cbuf)
	ld	e,(hl)
	inc	hl
	ld	d,(hl)			; de -> next chain to transmit
	inc	hl			; hl might be next cbuf entry pointer
	push	de
	ld	de,#tx_bottom
	or	a			;clear carry
	push	hl			;save what might be correct value
	sbc	hl,de
	pop	hl
	pop	de
        jr      nz,txn_1                ;go there if not at end of circ. buf

	ld	hl,#tx_top		;else we wrap aroune
txn_1:
	ld	(tx_head_cbuf),hl	;save next circ buf pointer in mem
	ex	de,hl			;return ptr to next chain to tx in hl
	ld	(tx_chain_head),hl	;tx rca routine needs this
	pop	de
	pop	af
	ret


